@@ -7,6 +7,7 @@ __author__ = ["mloning, fkiraly"]
 __all__ = ["_HeterogenousMetaEstimator"]
 
 from abc import ABCMeta
+from inspect import isclass
 
 from sklearn import clone
 
@@ -27,6 +28,19 @@ class _HeterogenousMetaEstimator(BaseEstimator, metaclass=ABCMeta):
         """Set estimator parameters."""
         raise NotImplementedError("abstract method")
 
+    def is_composite(self):
+        """Check if the object is composite.
+
+        A composite object is an object which contains objects, as parameters.
+        Called on an instance, since this may differ by instance.
+
+        Returns
+        -------
+        composite: bool, whether self contains a parameter which is BaseObject
+        """
+        # children of this class are always composite
+        return True
+
     def _get_params(self, attr, deep=True):
         out = super().get_params(deep=deep)
         if not deep:
@@ -87,7 +101,35 @@ class _HeterogenousMetaEstimator(BaseEstimator, metaclass=ABCMeta):
         subsetted_dict = dict((k, dict_to_subset[k]) for k in keys_in_both)
         return subsetted_dict
 
-    def _check_estimators(self, estimators, attr_name="steps", cls_type=None):
+    @staticmethod
+    def _is_name_and_est(obj, cls_type=None):
+        """Check whether obj is a tuple of type (str, cls_type).
+
+        Parameters
+        ----------
+        cls_type : class or tuple of class, optional. Default = BaseEstimator.
+            class(es) that all estimators are checked to be an instance of
+
+        Returns
+        -------
+        bool : True if obj is (str, cls_type) tuple, False otherise
+        """
+        if cls_type is None:
+            cls_type = BaseEstimator
+        if not isinstance(obj, tuple) or len(obj) != 2:
+            return False
+        if not isinstance(obj[0], str) or not isinstance(obj[1], cls_type):
+            return False
+        return True
+
+    def _check_estimators(
+        self,
+        estimators,
+        attr_name="steps",
+        cls_type=None,
+        allow_mix=True,
+        clone_ests=True,
+    ):
         """Check that estimators is a list of estimators or list of str/est tuples.
 
         Parameters
@@ -97,8 +139,12 @@ class _HeterogenousMetaEstimator(BaseEstimator, metaclass=ABCMeta):
             estimators should inherit from cls_type class
         attr_name : str, optional. Default = "steps"
             Name of checked attribute in error messages
-        cls_type : class, optional. Default = BaseEstimator.
-            class that all estimators are checked to be an instance of
+        cls_type : class or tuple of class, optional. Default = BaseEstimator.
+            class(es) that all estimators are checked to be an instance of
+        allow_mix : boolean, optional. Default = True.
+            whether mix of estimator and (str, estimator) is allowed in `estimators`
+        clone_ests : boolean, optional. Default = True.
+            whether estimators in return are cloned (True) or references (False).
 
         Returns
         -------
@@ -116,9 +162,15 @@ class _HeterogenousMetaEstimator(BaseEstimator, metaclass=ABCMeta):
             " of estimators, or a list of (string, estimator) tuples. "
         )
         if cls_type is None:
+            msg += f"All estimators in '{attr_name}' must be of type BaseEstimator."
             cls_type = BaseEstimator
+        elif isclass(cls_type) or isinstance(cls_type, tuple):
+            msg += (
+                f"All estimators in '{attr_name}' must be of type "
+                f"{cls_type.__name__}."
+            )
         else:
-            msg += f"All estimators must be of type {cls_type}."
+            raise TypeError("cls_type must be a class or tuple of classes")
 
         if (
             estimators is None
@@ -127,27 +179,60 @@ class _HeterogenousMetaEstimator(BaseEstimator, metaclass=ABCMeta):
         ):
             raise TypeError(msg)
 
-        if not isinstance(estimators[0], (cls_type, tuple)):
+        def is_est_is_tuple(obj):
+            """Check whether obj is estimator of right type, or (str, est) tuple."""
+            is_est = isinstance(obj, cls_type)
+            is_tuple = self._is_name_and_est(obj, cls_type)
+
+            return is_est, is_tuple
+
+        if not all(any(is_est_is_tuple(x)) for x in estimators):
             raise TypeError(msg)
 
-        if isinstance(estimators[0], cls_type):
-            if not all(isinstance(est, cls_type) for est in estimators):
-                raise TypeError(msg)
-        if isinstance(estimators[0], tuple):
-            if not all(isinstance(est, tuple) for est in estimators):
-                raise TypeError(msg)
-            if not all(isinstance(est[0], str) for est in estimators):
-                raise TypeError(msg)
-            if not all(isinstance(est[1], cls_type) for est in estimators):
-                raise TypeError(msg)
+        msg_no_mix = (
+            f"elements of {attr_name} must either all be estimators, "
+            f"or all (str, estimator) tuples, mix of the two is not allowed"
+        )
 
-        return self._get_estimator_tuples(estimators, clone_ests=True)
+        if not allow_mix and not all(is_est_is_tuple(x)[0] for x in estimators):
+            if not all(is_est_is_tuple(x)[1] for x in estimators):
+                raise TypeError(msg_no_mix)
+
+        return self._get_estimator_tuples(estimators, clone_ests=clone_ests)
+
+    def _coerce_estimator_tuple(self, obj, clone_est=False):
+        """Coerce estimator or (str, estimator) tuple to (str, estimator) tuple.
+
+        Parameters
+        ----------
+        obj : estimator or (str, estimator) tuple
+            assumes that this has been checked, no checks are performed
+        clone_est : boolean, optional. Default = False.
+            Whether to return clone of estimator in obj (True) or a reference (False).
+
+        Returns
+        -------
+        est_tuple : (str, stimator tuple)
+            obj if obj was (str, estimator) tuple
+            (obj class name, obj) if obj was estimator
+        """
+        if isinstance(obj, tuple):
+            est = obj[1]
+            name = obj[0]
+        else:
+            est = obj
+            name = type(obj).__name__
+
+        if clone_est:
+            return (name, clone(est))
+        else:
+            return (name, est)
 
     def _get_estimator_list(self, estimators):
         """Return list of estimators, from a list or tuple.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         estimators : list of estimators, or list of (str, estimator tuples)
 
         Returns
@@ -155,16 +240,13 @@ class _HeterogenousMetaEstimator(BaseEstimator, metaclass=ABCMeta):
         list of estimators - identical with estimators if list of estimators
             if list of (str, estimator) tuples, the str get removed
         """
-        if isinstance(estimators[0], tuple):
-            return [x[1] for x in estimators]
-        else:
-            return estimators
+        return [self._coerce_estimator_tuple(x)[1] for x in estimators]
 
     def _get_estimator_names(self, estimators, make_unique=False):
         """Return names for the estimators, optionally made unique.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         estimators : list of estimators, or list of (str, estimator tuples)
         make_unique : bool, optional, default=False
             whether names should be made unique in the return
@@ -175,18 +257,7 @@ class _HeterogenousMetaEstimator(BaseEstimator, metaclass=ABCMeta):
             names for estimators in estimators
             if make_unique=True, made unique using _make_strings_unique
         """
-        if estimators is None or len(estimators) == 0:
-            names = []
-        elif isinstance(estimators[0], tuple):
-            names = [x[0] for x in estimators]
-        elif isinstance(estimators[0], BaseEstimator):
-            names = [type(e).__name__ for e in estimators]
-        else:
-            raise RuntimeError(
-                "unreachable condition in _get_estimator_names, "
-                " likely input assumptions are violated,"
-                " run _check_estimators before running _get_estimator_names"
-            )
+        names = [self._coerce_estimator_tuple(x)[0] for x in estimators]
         if make_unique:
             names = self._make_strings_unique(names)
         return names
@@ -194,8 +265,8 @@ class _HeterogenousMetaEstimator(BaseEstimator, metaclass=ABCMeta):
     def _get_estimator_tuples(self, estimators, clone_ests=False):
         """Return list of estimator tuples, from a list or tuple.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         estimators : list of estimators, or list of (str, estimator tuples)
         clone_ests : bool, whether estimators get cloned in the process
 
@@ -292,8 +363,8 @@ class _HeterogenousMetaEstimator(BaseEstimator, metaclass=ABCMeta):
         """Set self's `tag_name` tag to `value` if any estimator on the list has it.
 
         Writes to self:
-        tag with name tag_name, sets to value if _anytagis(tag_name, value) is True
-            otherwise sets the tag to `value_if_not`
+        sets the tag `tag_name` to `value` if `_anytagis(tag_name, value)` is True
+            otherwise sets the tag `tag_name` to `value_if_not`
 
         Parameters
         ----------
@@ -340,6 +411,88 @@ class _HeterogenousMetaEstimator(BaseEstimator, metaclass=ABCMeta):
         if tag_val != "None":
             self.set_tags(**{tag_name: tag_val})
 
+    def _tagchain_is_linked(
+        self,
+        left_tag_name,
+        mid_tag_name,
+        estimators,
+        left_tag_val=True,
+        mid_tag_val=True,
+    ):
+        """Check whether all tags left of the first mid_tag/val are left_tag/val.
+
+        Useful to check, for instance, whether all instances of estimators
+            left of the first missing value imputer can deal with missing values.
+
+        Parameters
+        ----------
+        left_tag_name : str, name of the left tag
+        mid_tag_name : str, name of the middle tag
+        estimators : list of (str, estimator) pairs to query for the tag/value
+        left_tag_val : value of the left tag, optional, default=True
+        mid_tag_val : value of the middle tag, optional, default=True
+
+        Returns
+        -------
+        chain_is_linked : bool,
+            True iff all "left" tag instances `left_tag_name` have value `left_tag_val`
+            a "left" tag instance is an instance in estimators which is earlier
+            than the first occurrence of `mid_tag_name` with value `mid_tag_val`
+        chain_is_complete : bool,
+            True iff chain_is_linked is True, and
+                there is an occurrence of `mid_tag_name` with value `mid_tag_val`
+        """
+        for _, est in estimators:
+            if est.get_tag(mid_tag_name) == mid_tag_val:
+                return True, True
+            if not est.get_tag(left_tag_name) == left_tag_val:
+                return False, False
+        return True, False
+
+    def _tagchain_is_linked_set(
+        self,
+        left_tag_name,
+        mid_tag_name,
+        estimators,
+        left_tag_val=True,
+        mid_tag_val=True,
+        left_tag_val_not=False,
+        mid_tag_val_not=False,
+    ):
+        """Check if _tagchain_is_linked, then set self left_tag_name and mid_tag_name.
+
+        Writes to self:
+        tag with name left_tag_name, sets to left_tag_val if _tag_chain_is_linked[0]
+            otherwise sets to left_tag_val_not
+        tag with name mid_tag_name, sets to mid_tag_val if _tag_chain_is_linked[1]
+            otherwise sets to mid_tag_val_not
+
+        Parameters
+        ----------
+        left_tag_name : str, name of the left tag
+        mid_tag_name : str, name of the middle tag
+        estimators : list of (str, estimator) pairs to query for the tag/value
+        left_tag_val : value of the left tag, optional, default=True
+        mid_tag_val : value of the middle tag, optional, default=True
+        left_tag_val_not : value to set if not linked, optional, default=False
+        mid_tag_val_not : value to set if not linked, optional, default=False
+        """
+        linked, complete = self._tagchain_is_linked(
+            left_tag_name=left_tag_name,
+            mid_tag_name=mid_tag_name,
+            estimators=estimators,
+            left_tag_val=left_tag_val,
+            mid_tag_val=mid_tag_val,
+        )
+        if linked:
+            self.set_tags(**{left_tag_name: left_tag_val})
+        else:
+            self.set_tags(**{left_tag_name: left_tag_val_not})
+        if complete:
+            self.set_tags(**{mid_tag_name: mid_tag_val})
+        else:
+            self.set_tags(**{mid_tag_name: mid_tag_val_not})
+
 
 def flatten(obj):
     """Flatten nested list/tuple structure.
